#!/usr/bin/python

"""
2008-WebCrawler.

Usage:
    crawler --feed=<url>
    crawler --help

Options:
    -h --help    Show this screen.
    --feed=<url> Web site to crawl.

"""

from collections import deque
import datetime
from docopt import docopt
import sys
import time
import traceback
import urllib.request
import urllib.error
import urllib.parse

from pycrawlerlib.entities import UrlFilter
from pycrawlerlib.factories import web_page_from_url


class UrlFrontier(deque):
    """
    Contains urls wet to be visited.
    """

    previous_url_count = 0

    def __str__(self):
        return str(len(self))

    def unchanged(self) -> bool:
        """
        Checks if the number of urls remains unchanged.
        If the number of urls remains constant,
         it may be that the crawler is stalled.
        """

        if len(self) == self.previous_url_count:
            return True

        self.previous_url_count = len(self)
        return False


class Crawler(object):
    """
    A Crawler instance visits an web page taking into account robot.txt rules.
    """

    def __init__(self, feed):

        self.url_frontier = UrlFrontier()
        self.content_seen = dict()  # Stores an hash of the web pages already visited.
        self.urls_seen = dict()  # Urls already visited
        self.url_filter = UrlFilter()

        self.working_url = ""
        self.feed = feed

        # Blacklist variables
        self.black_list_dict = {}
        self.max_sequential_access = 0
        self.number_of_sequential_access = 0
        self.black_list_timeout = 0

        self.url_frontier.appendleft(feed)

    def start(self):
        """
        Starts the web crawler main loop.
        """

        print("CRAWLER: starting crawler")

        previous_url = ""
        while True:
            # Fetch an url to work on
            try:
                self.working_url = self.url_frontier.pop()
            except IndexError:
                print("CRAWLER: no more urls to test")
                return

            # Ignore sites outside current site
            if not self.working_url.startswith(self.feed):
                self.urls_seen[self.working_url] = 0
                continue

            print("\nCRAWLER: working on url %s" % self.working_url)

            # Check ir url was already seen
            try:
                self.urls_seen[self.working_url]
            except KeyError:
                pass
            else:
                print("CRAWLER: already seen")
                continue

            # Check if url is blacklisted
            if self.blacklisted(self.working_url, previous_url):
                print("CRAWLER: url download not allowed")
                self.url_frontier.appendleft(self.working_url)
                continue

            web_page = web_page_from_url(self.working_url)
            if web_page:
                # Check if page content was already seen
                try:
                    self.content_seen[web_page.hash]
                except KeyError:
                    pass
                else:
                    print("CRAWLER: page already seen")
                    continue

                # Add page urls to url frontier
                for url in web_page:
                    try:
                        self.urls_seen[url]
                    except KeyError:
                        if self.url_filter.can_crawl(url):
                            self.url_frontier.appendleft(url)

                self.urls_seen[self.working_url] = datetime.datetime.now()
                self.content_seen[web_page.hash] = datetime.datetime.now()
                print("CRAWLER: ", self.working_url)

            previous_url = self.working_url

            if self.url_frontier.unchanged():
                print("WARNING: crawler may be stalled")

    def blacklisted(self, url: str, url_anterior: str) -> bool:
        """
        Avoids sequential access to a page in a short amount of space.
        """

        netloc_anterior = (urllib.parse.urlparse(url_anterior))[1]
        netloc_actual = (urllib.parse.urlparse(url))[1]

        if netloc_actual in self.black_list_dict:
            temp_lock = time.time() - self.black_list_dict[netloc_actual]

            if temp_lock > self.black_list_timeout:
                print(netloc_actual, " was eliminated from blacklist after ", temp_lock, " seconds")
                del self.black_list_dict[netloc_actual]
                return False

            print(netloc_actual, " is in the blacklist")
            return True

        if netloc_actual == netloc_anterior:
            self.number_of_sequential_access = self.number_of_sequential_access + 1
        else:
            self.number_of_sequential_access = 0

        if self.number_of_sequential_access > self.max_sequential_access:
            print("adding ", netloc_actual, " to the blacklist due to excess of sequential accesses")
            self.black_list_dict[netloc_actual] = time.time()
            self.number_of_sequential_access = 0
            return True

        return False

    def configuration(self):
        self.max_sequential_access = 2
        self.black_list_timeout = 20


if __name__ == "__main__":
    arguments = docopt(__doc__, help=True)
    crawler = Crawler(arguments['--feed'])
    crawler.configuration()

    try:
        crawler.start()
    except Exception as e:
        print(traceback.print_tb(e.__traceback__))
        print(str(e))
        sys.exit(1)
