import urllib
import urllib.parse
from hashlib import md5

from bs4 import BeautifulSoup

from pycrawlerlib.robot_parser import RobotExclusionRulesParser
import requests


class UrlFilter(object):
    def __init__(self):
        self.robots_txt_cache = {}
        self.rules = RobotExclusionRulesParser()

    def can_crawl(self, url) -> bool:
        """
        Checks if a page can be crawled.
        """

        up = urllib.parse.urlparse(url)
        authority = up[1]
        schema = up[0]
        root = schema + "://" + authority + "/"

        if not authority:
            return False

        try:
            self.robots_txt_cache[authority]
        except KeyError:
            robots_txt_content = self._download_content(root + "robots.txt")
            if not robots_txt_content:
                return False

            self.robots_txt_cache[authority] = robots_txt_content

        self.rules.parse(self.robots_txt_cache[authority])
        if self.rules.is_expired():
            robots_txt_content = self._download_content(root + "robots.txt")
            if not robots_txt_content:
                return False

            self.robots_txt_cache[authority] = robots_txt_content

        try:
            is_allowed = self.rules.is_allowed('*', url)
        except Exception as e:
            print("error in url filter: %s" % str(e))
            return False

        if is_allowed:
            return True

    @staticmethod
    def _download_content(url: str) -> str:
        """
        Downloads the page identified by the url.
        """

        try:
            response = requests.get(url, timeout=5)
        except requests.exceptions.ConnectionError:
            pass
        except requests.exceptions.ReadTimeout:
            pass
        else:
            if response.status_code == 200:
                return response.content.decode()


class WebPage(object):
    """
    An WebPage contains urls and it's an url iterator.
    """

    html: str
    url: str
    hash = None

    def __init__(self, html: str, url: str):
        self.html = html

        m = md5()
        m.update(self.html.encode('utf-8'))
        self.hash = m.hexdigest()

        self.url = url
        self.urls = []
        self._urls_for_page()

        self.current_url_idx = 0

    def __len__(self):
        return len(self.urls)

    def __iter__(self):
        return self

    def __next__(self):
        if self.current_url_idx == len(self) - 1:
            raise StopIteration
        else:
            self.current_url_idx += 1
            return self.urls[self.current_url_idx]

    # TODO: url show be a property
    def _urls_for_page(self):
        """
        Urls contained in the web page.
        """

        # TODO: Make this run faster. Use itertools. Do benchmarks
        # TODO: Confirm in someway all urls are being fetched from the url

        if self.urls:
            return self.urls

        urls = []
        html_soup = BeautifulSoup(self.html, features="html.parser")
        for url in html_soup.find_all('a'):
            urls.append(url.get('href'))

        urls = [u for u in urls if u and u != "/"]
        urls = [u for u in urls if u and u.startswith('/') or u.startswith('http')]

        aux_urls = []
        if urls:
            for i, url in enumerate(urls):
                aux_urls.append(url)
                if url.startswith('/'):
                    up = urllib.parse.urlparse(self.url)
                    authority = up[1]
                    schema = up[0]
                    aux_urls[i] = schema + "://" + authority + url

                self.urls = aux_urls

        return self.urls
